@@ -22,8 +22,8 @@ except ImportError:
     sys.exit(1)
 
 from rdflib import Graph, Dataset
-from rdflib.term import Literal, URIRef, Variable
-from rdflib.namespace import Namespace, RDF, RDFS
+from rdflib.term import Node, Literal, URIRef, Variable
+from rdflib.namespace import Namespace, RDF, RDFS, OWL
 
 has_reasoner = False
 try:
@@ -75,7 +75,7 @@ N3_PROLOGUE += (
 from .exceptions import KbServerError
 from minimalkb import __version__
 
-from .services import lifespan
+#from .services import lifespan
 
 from .helpers import memoize
 
@@ -162,6 +162,12 @@ def shorten_graph(graph):
 def get_variables(stmt):
     return [v for v in stmt if isinstance(v, Variable)]
 
+def get_all_variables(stmts):
+    vars = []
+    for p in stmts:
+       vars += [v.n3() for v in get_variables(parse_stmt(p))]
+
+    return vars
 
 class dotdict(dict):
     """dot.notation access to dictionary attributes"""
@@ -171,54 +177,38 @@ class dotdict(dict):
     __delattr__ = dict.__delitem__
 
 
+class hashabledict(dict):
+    def __hash__(self):
+        return hash(tuple(sorted(self.items())))
+
+
 class Event:
-
-    NEW_INSTANCE = "NEW_INSTANCE"
-    NEW_CLASS_INSTANCE = "NEW_CLASS_INSTANCE"
-    NEW_INSTANCE_ONE_SHOT = "NEW_INSTANCE_ONE_SHOT"
-    NEW_CLASS_INSTANCE_ONE_SHOT = "NEW_CLASS_INSTANCE_ONE_SHOT"
-
-    def __init__(self, kb, type, trigger, var, patterns, models):
+    def __init__(self, kb, patterns, one_shot, models):
         self.kb = kb
 
-        if type == Event.NEW_CLASS_INSTANCE:
-            self.type = Event.NEW_INSTANCE
-            self.var = "?instance"
-            self.patterns = ["?instance rdf:type %s" % klass[0] for klass in patterns]
-        elif type == Event.NEW_CLASS_INSTANCE_ONE_SHOT:
-            self.type = Event.NEW_INSTANCE_ONE_SHOT
-            self.var = "?instance"
-            self.patterns = ["?instance rdf:type %s" % klass[0] for klass in patterns]
-        else:
-            self.type = type
-            self.var = var
-            self.patterns = patterns
+        self.patterns = patterns
+        self.one_shot = one_shot
 
-        self.trigger = trigger
         self.models = models
 
         self.id = "evt_" + str(
             stable_hash(
-                self.type
-                + self.trigger
-                + (self.var if self.var else "")
-                + str(sorted(self.patterns))
-                + str(sorted(self.models))
+                str(sorted(self.patterns)) + str(one_shot) + str(sorted(self.models))
             )
         )
 
         self.content = None
 
+        self.vars = get_all_variables(self.patterns)
+
         self.valid = True
 
         self.previous_instances = set()
-        if type in [Event.NEW_INSTANCE, Event.NEW_INSTANCE_ONE_SHOT]:
 
-            instances = self.kb.find([self.var], self.patterns, frozenset(self.models))
-            logger.debug(
-                "Creating a NEW_INSTANCE event with initial instances %s" % instances
-            )
-            self.previous_instances = set(instances)
+        instances = self.kb.find( self.patterns, self.vars, frozenset(self.models))
+        logger.debug("Creating a event with initial instances %s" % instances)
+
+        self.previous_instances = set([hashabledict(row) for row in instances])
 
     def __hash__(self):
         return hash(self.id)
@@ -227,27 +217,28 @@ class Event:
         return hash(self).__cmp__(hash(other))
 
     def evaluate(self):
-        if "ONE_SHOT" in self.trigger:
+        if not self.valid:
+            return False
+
+        if self.one_shot:
             self.valid = False
 
-        if self.type in [Event.NEW_INSTANCE, Event.NEW_INSTANCE_ONE_SHOT]:
-            instances = set(
-                self.kb.find([self.var], self.patterns, frozenset(self.models))
-            )
-            newinstances = instances - self.previous_instances
+        instances = self.kb.find(self.patterns, self.vars, frozenset(self.models))
 
-            # previous_instances must be set to the current set of matching instance
-            # else we won't trigger events when an instance disappear and
-            # re-appear later.
-            self.previous_instances = instances
+        instances = set([hashabledict(row) for row in instances])
 
-            if not newinstances:
-                return False
+        newinstances = instances - self.previous_instances
 
-            self.content = [
-                i for i in newinstances
-            ]  # for some reason, calling list() does not work
-            return True
+        # previous_instances must be set to the current set of matching instance
+        # else we won't trigger events when an instance disappear and
+        # re-appear later.
+        self.previous_instances = instances
+
+        if not newinstances:
+            return False
+
+        self.content = list(newinstances)
+        return True
 
 
 class MinimalKB:
@@ -308,6 +299,9 @@ class MinimalKB:
         for model in models:
             logger.info("Loading file <%s> in model <%s>" % (filename, model))
             self.models[model].graph.parse(filename, publicID=IRIS[DEFAULT_PREFIX])
+            self.models[model].is_dirty = True
+
+        self.onupdate()
 
     @api
     def clear(self):
@@ -324,27 +318,45 @@ class MinimalKB:
         return list(self._api.keys())
 
     @api
-    def about(self, term, models=None):
+    def about(self, raw_term, models=None):
 
         result = []
 
         models = self.normalize_models(models)
 
-        term = parse_term(term)
+        term = None
+        try:
+            term = parse_term(raw_term)
+        except KbServerError:
+            # error while parsing the raw term: probably a
+            # string with spaces. In that case, only search in the labels.
+            pass
 
         for model in models:
             g = self.models[model].materialized_graph
-            for triple in g.triples((term, None, None)):
-                result.append(shorten(g, triple))
-            for triple in g.triples((None, term, None)):
-                result.append(shorten(g, triple))
-            for triple in g.triples((None, None, term)):
-                result.append(shorten(g, triple))
+
+            if term:
+                for triple in g.triples((term, None, None)):
+                    result.append(shorten(g, triple))
+                for triple in g.triples((None, term, None)):
+                    result.append(shorten(g, triple))
+                for triple in g.triples((None, None, term)):
+                    result.append(shorten(g, triple))
+
+            for s, p, o in g.triples((None, RDFS.label, None)):
+                if raw_term in o.value:
+                    result.append(shorten(g, (s, p, o)))
 
         return result
 
     @api
     def lookup(self, resource, models=None):
+        """Search the knowledge base for a term matching a string. The search is
+        performed both on terms' names and on label.
+
+        Returns the list of found terms, alongside with their type (one of instance, class,
+        datatype_property, object_property, literal)
+        """
         models = self.normalize_models(models)
         logger.info(
             "Lookup for "
@@ -363,25 +375,26 @@ class MinimalKB:
         for s, p, o in about:
             if s == resource or p == resource or o == resource:
                 matching_concepts.add(resource)
+            elif p == shorten_term(self.models[DEFAULT_MODEL].graph, RDFS.label):
+                matching_concepts.add(s)
 
         res = [(concept, self.typeof(concept, models)) for concept in matching_concepts]
         logger.info("Found: " + str(res))
         return res
 
-    @api
     def _instancesof(self, term, direct=False, models=[]):
 
-        if direct:
-            logger.warn(
-                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
-            )
-
         models = self.normalize_models(models)
         result = []
 
-        res = parse_term(term)
+        if not isinstance(term, Node):
+            term = parse_term(term)
+
         for model in models:
-            result += self.models[model].materialized_graph.subjects(RDF.type, res)
+            if direct:
+                result += self.models[model].graph.subjects(RDF.type, term)
+            else:
+                result += self.models[model].materialized_graph.subjects(RDF.type, term)
 
         return result
 
@@ -394,80 +407,110 @@ class MinimalKB:
 
     def _classesof(self, term, direct=False, models=[]):
 
-        if direct:
-            logger.warn(
-                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
-            )
-
         models = self.normalize_models(models)
         result = []
 
-        res = parse_term(term)
+        if not isinstance(term, Node):
+            term = parse_term(term)
+
         for model in models:
-            result += self.models[model].materialized_graph.objects(res, RDF.type)
+            if direct:
+                result += self.models[model].graph.objects(term, RDF.type)
+            else:
+                result += self.models[model].materialized_graph.objects(term, RDF.type)
 
         return result
 
-    @api
     def _subclassesof(self, term, direct=False, models=[]):
 
-        if direct:
-            logger.warn(
-                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
-            )
+        models = self.normalize_models(models)
+        result = []
+
+        if not isinstance(term, Node):
+            term = parse_term(term)
+
+        for model in models:
+            if direct:
+                result += self.models[model].graph.subjects(RDFS.subClassOf, term)
+            else:
+                result += self.models[model].materialized_graph.subjects(
+                    RDFS.subClassOf, term
+                )
+
+        return result
+
+    def _superclassesof(self, term, direct=False, models=[]):
 
         models = self.normalize_models(models)
         result = []
 
-        res = parse_term(term)
+        if not isinstance(term, Node):
+            term = parse_term(term)
+
         for model in models:
-            result += self.models[model].materialized_graph.subjects(
-                RDFS.subClassOf, res
-            )
+            if direct:
+                result += self.models[model].graph.objects(term, RDFS.subClassOf)
+            else:
+                result += self.models[model].materialized_graph.objects(
+                    term, RDFS.subClassOf
+                )
 
         return result
 
     @api
-    def _superclassesof(self, term, direct=False, models=[]):
-
-        if direct:
-            logger.warn(
-                "'direct=True' not implemented! Returing all asserted 'is-a' relations."
-            )
-
+    def label(self, term, models=[]):
+        """returns the labels attached to a term, as a dictionary `{"default":"label1", "lang_code1": label1,
+        "lang_code2": "label2",...}` where the 'default' key returns either the
+        English version of the label, or the name of the term, if no label is
+        available, and the other keys provide localised version of the label, if
+        available in the knowledge base.
+        """
         models = self.normalize_models(models)
-        result = []
+        result = {}
+
+        if not isinstance(term, URIRef):
+            term = parse_term(term)
 
-        res = parse_term(term)
         for model in models:
-            result += self.models[model].materialized_graph.objects(
-                res, RDFS.subClassOf
-            )
+
+            labels = self.models[model].materialized_graph.objects(term, RDFS.label)
+            for label in labels:
+                if label.language:
+                    result[label.language] = label
+                else:
+                    result["default"] = label
+
+        if not "default" in result:
+            if "en" in result:
+                result["default"] = result["en"]
+            else:
+                result["default"] = shorten_term(self.models[DEFAULT_MODEL].graph, term)
 
         return result
 
-    def typeof(self, raw_term, models):
+    def typeof(self, term, models):
 
-        term = parse_term(raw_term)
+        if not isinstance(term, Node):
+            term = parse_term(term)
 
         if isinstance(term, Literal):
             return "literal"
 
-        classes = self._classesof(raw_term, False, models)
+        classes = self._classesof(term, False, models)
         if classes:
-            if "owl:ObjectProperty" in classes:
+            if OWL.ObjectProperty in classes:
                 return "object_property"
-            elif "owl:DatatypeProperty" in classes:
+            elif OWL.DatatypeProperty in classes:
                 return "datatype_property"
-            elif "owl:Class" in classes:
+            elif OWL.Class in classes:
                 return "class"
             else:
                 return "instance"
 
         if (
-            self._instancesof(raw_term, False, models)
-            or self._subclassesof(raw_term, False, models)
-            or self._superclassesof(raw_term, False, models)
+            self._instancesof(term, False, models)
+            or self._subclassesof(term, False, models)
+            or self._superclassesof(term, False, models)
         ):
             return "class"
 
@@ -489,7 +532,7 @@ class MinimalKB:
     def details(self, resource, models=None):
         """
         Returns a dictionary containing the following details on a given resource:
-        - 'name': resource label, if any, literal value for literals, else resource ID.
+        - 'label': resource label, if any, literal value for literals, else resource ID.
         - 'id': resource ID or 'literal' for literals
         - 'type': one of ['instance', 'class', 'object_property', 'datatype_property', 'undecided']
         - 'sameAs': list of equivalent classes or instances, if any
@@ -505,9 +548,9 @@ class MinimalKB:
         """
         models = self.normalize_models(models)
         res = {}
-        res["name"] = self.store.label(resource, models)
+        res["label"] = self.label(resource, models)
         res["id"] = resource
-        res["type"] = self.store.typeof(resource, models)
+        res["type"] = self.typeof(resource, models)
 
         if res["type"] == "class":
             res["attributes"] = []
@@ -516,7 +559,7 @@ class MinimalKB:
                     "name": "Parents",
                     "id": "superClasses",
                     "values": [
-                        {"id": r, "name": self.store.label(r, models)}
+                        {"id": r, "name": self.label(r, models)}
                         for r in self._superclassesof(resource, True, models)
                     ],
                 }
@@ -527,7 +570,7 @@ class MinimalKB:
                     "name": "Children",
                     "id": "subClasses",
                     "values": [
-                        {"id": r, "name": self.store.label(r, models)}
+                        {"id": r, "label": setermlabel(r, models)}
                         for r in self._subclassesof(resource, True, models)
                     ],
                 }
@@ -538,7 +581,7 @@ class MinimalKB:
                     "name": "Instances",
                     "id": "instances",
                     "values": [
-                        {"id": r, "name": self.store.label(r, models)}
+                        {"id": r, "label": self.label(r, models)}
                         for r in self._instancesof(resource, True, models)
                     ],
                 }
@@ -550,22 +593,13 @@ class MinimalKB:
                     "name": "Classes",
                     "id": "classes",
                     "values": [
-                        {"id": r, "name": self.store.label(r, models)}
-                        for r in self.store._classesof(resource, True, models)
+                        {"id": r, "label": self.label(r, models)}
+                        for r in self._classesof(resource, True, models)
                     ],
                 }
             ]
         return res
 
-    @api
-    def check(self, stmts, models=None):
-        logger.warn(
-            "'check' has been invoked, but no classification supported."
-            "Returning true if the statements are already asserted/inferred,"
-            "false otherwise (ie, calls 'exist' instead of 'check')."
-        )
-        return self.exist(stmts, models)
-
     @api
     def exist(self, raw_stmts, models=None):
         """Returns True if all the statements exist (eg, are materialised) in
@@ -582,9 +616,7 @@ class MinimalKB:
         )
         stmts = [parse_stmt(s) for s in raw_stmts]
 
-        vars = []
-        for stmt in stmts:
-            vars += [v.n3() for v in get_variables(stmt)]
+        vars = get_all_variables(raw_stmts)
 
         # no variable? simply check if all the statements are asserted
         if len(vars) == 0:
@@ -596,7 +628,7 @@ class MinimalKB:
             return True
 
         # else, run a query
-        return bool(self.find(vars, raw_stmts, models))
+        return bool(self.find(raw_stmts, vars, models))
 
     @api
     def revise(self, stmts, policy):
@@ -609,14 +641,15 @@ class MinimalKB:
           seconds, float
         """
 
+        if type(policy) != dict:
+            raise KbServerError("Expected a dictionary as policy")
+
         if isinstance(stmts, str):
             raise KbServerError("A list of statements is expected")
 
         subgraph = parse_stmts_to_graph(stmts)
         parsed_stmts = "\n\t- ".join([" ".join(s) for s in shorten_graph(subgraph)])
 
-        if type(policy) != dict:
-            raise KbServerError("Expected a dictionary as policy")
         models = self.normalize_models(policy.get("models", []))
 
         if policy["method"] in ["add", "safe_add"]:
@@ -725,30 +758,43 @@ class MinimalKB:
         return json.loads(sparql_res.serialize(format="json"))
 
     @api
-    def find(self, vars, patterns, models=None):
+    def find(self, patterns, vars=None, models=None):
         """
-        Depending on the arguments, three differents
-        behaviours are possible:
+        Performs a query on one or several models.
 
-        - if len(vars) == 0, 'find' returns the statement/set of statements if they all exist in the model.
-        - if len(vars) == 1, 'find' returns a set of resources matching the patterns.
-        - if len(vars) > 1:
-            - if len(patterns) == 1, a list of statements matching the pattern
-              is returned.
-            - else, a list of dictionaries is returned with
-            possible combination of values for the different variables. For
-            instance, find(["?agent", "?action"], ["?agent desires ?action", "?action rdf:type Jump"])
-            would return something like: [{"agent":"james", "action": "jumpHigh"}, {"agent": "laurel", "action":"jumpHigher"}]
+        `patterns`: a list of triples containing unbound terms (eg, variables).
+        As a convenience, if the patterns do not contain any variable, `find`
+        will return whether the provided list of triples is present in the
+        model(s).
+
+        `vars`: the list of variables (prefixed with `?`) that you want to
+        select. If None or an empty list, returns all the variables found in
+        the patterns.
+
+        `models`: the list of models you want to run the query against. The
+        returned result will be the conjunction of results when running the
+        query on each models separately.
+
+        A list of dictionaries is returned with possible combination of values
+        for the different variables. For instance, find(["?agent", "?action"],
+        ["?agent desires ?action", "?action rdf:type Jump"]) would return
+        something like: [{"agent":"james", "action": "jumpHigh"}, {"agent":
+        "laurel", "action":"jumpHigher"}]
+
+        If you were using anonymous variables (starting with ?__) in your query
+        (eg generated by `pykb` when using a `*` wildcard), they will be
+        renamed `var1`, `var2`, etc.
         """
 
         if not vars:
-            if not self.exist(patterns, models):
-                return []
-            return patterns
+            vars = get_all_variables(patterns)
+            if len(vars) == 0:
+                return self.exist(patterns, models)
 
         models = self.normalize_models(models)
 
         patterns = [parse_stmt(p) for p in patterns]
+
         parsed_patterns = "\n\t- ".join(
             [" ".join(s) for s in shortenN(self.models[DEFAULT_MODEL].graph, patterns)]
         )
@@ -767,13 +813,19 @@ class MinimalKB:
         res = []
 
         named_variables = self.named_variables(vars)
-        # if we have named variable + extra 'anonymous' variable (coming from
-        # eg '*'), ignore them
-        if named_variables and len(named_variables) != len(vars):
-            vars = named_variables
+        if named_variables:
+            # if we have named variable + extra 'anonymous' variable (coming from
+            # eg '*'), ignore the anonymous ones
+            if len(named_variables) != len(vars):
+                vars = named_variables
+
+            # remove the '?' at the start of the variables
+            vars_naked = [var[1:] for var in vars]
+
+        # only anonymous variables? rename them var1,var2,var3,var4...
+        else:
+            vars_naked = ["var%s" % (i + 1) for i in range(len(vars))]
 
-        # remove the '?' at the start of the variables
-        vars_naked = [var[1:] for var in vars]
 
         for model in models:
             q = "SELECT %s WHERE {\n" % " ".join(vars)
@@ -783,36 +835,51 @@ class MinimalKB:
 
             sparql_res = self._sparql(model, q)
 
-            if len(named_variables) > 1:
-                res += [
-                    dict(zip(vars_naked, r))
-                    for r in [
-                        shorten(self.models[model].graph, row) for row in sparql_res
-                    ]
+            res += [
+                dict(zip(vars_naked, r))
+                for r in [
+                    shorten(self.models[model].graph, row) for row in sparql_res
                 ]
-            else:
-                res += [shorten(self.models[model].graph, row) for row in sparql_res]
-
-        # if a single variable is requested, 'unpack' the result
-        if len(vars) == 1:
-            res = [r[0] for r in res]
+            ]
 
         logger.info("Found: " + str(res))
         return res
 
     @api
-    def subscribe(self, type, trigger, var, patterns, models=None):
+    def subscribe(self, patterns, one_shot=False, models=None):
+        """Subscribes to a specified event in the ontology.
+
+        Every time the model(s) is(are) updated, the provided `patterns` are
+        evaluated against the set of asserted and inferred triples. If at least
+        one triple is returned, the event is fired.
+
+        The terms bounded to the named variables in the patterns are attached to
+        the fired event.
+
+        For instance:
+        ```
+        from kb import KB
+
+        def on_new_robot_instance(instances):
+            print("New robots: " + ", ".join(instances))
+
+        with KB() as kb:
+            kb.subscribe(["?robot rdf:type Robot"])
+
+            kb += ["myself rdf:type Robot"] # should print "New robots: myself"
+            time.sleep()
+        ```
+
+        If `one_shot` is set to true, the event is discarded once it has fired
+        once.
+
+        """
 
         models = self.normalize_models(models)
 
-        logger.info(
-            "Registering a new event: %s %s for %s on %s"
-            % (type, trigger, var, patterns)
-            + " in "
-            + (str(models) if models else "default model.")
-        )
+        logger.info("Registering a new event: %s" % patterns + " in " + str(models))
 
-        event = Event(self, type, trigger, var, patterns, models)
+        event = Event(self, patterns, one_shot, models)
 
         self.active_evts.add(event)
 
@@ -862,7 +929,7 @@ class MinimalKB:
             self._instancesof("owl:FunctionalProperty", False)
         )
 
-        for e in self.active_evts:
+        for e in list(self.active_evts):
             if e.evaluate():
                 clients = self.eventsubscriptions[e.id]
                 logger.info(
@@ -872,7 +939,8 @@ class MinimalKB:
                     msg = ("event", e)
                     self.requestresults.setdefault(client, Queue()).put(msg)
                 if not e.valid:
-                    self.active_evts.discard(e)
+                    self.active_evts.remove(e)
+                    del self.eventsubscriptions[e.id]
 
     def materialise(self, models=None):
 
@@ -950,6 +1018,8 @@ class MinimalKB:
                 if isinstance(models, str):
                     models = [models]
 
+                models = [m if m != "" else DEFAULT_MODEL for m in models]
+
                 # do we have a new model? initialise it.
                 for model in models:
                     if model not in self.models:
@@ -962,6 +1032,30 @@ class MinimalKB:
     def execute(self, client, name, *args, **kwargs):
 
         if name == "close":
+
+            ### Event-related housekepping
+            evts_id_to_remove = []
+            for evt_id, clients in self.eventsubscriptions.items():
+                if client in clients:
+                    logger.info(
+                        "Client closing, removing event subscription <%s>" % evt_id
+                    )
+                    clients.remove(client)
+                if len(clients) == 0:
+                    evts_id_to_remove.append(evt_id)
+            for evt_id in evts_id_to_remove:
+                logger.debug(
+                    "No one interested in event %s anymore. Removing it" % evt_id
+                )
+                del self.eventsubscriptions[evt_id]
+                evts_to_remove = []
+                for e in self.active_evts:
+                    if e.id == evt_id:
+                        evts_to_remove.append(e)
+                for e in evts_to_remove:
+                    self.active_evts.remove(e)
+            ###############
+
             logger.info("Closing connection to client.")
             return
 
